/***************************************************************************
  tag: Rock Chen  Wednesday March 25 2015,  main.cpp

                        main.cpp -  description
                           -------------------
    begin                : March 25 2015
    copyright            : (C) 2015 Rock Chen
    email                : rockcdr@bz-intelligentrobot.com

******************************************************************************/
 /*****************************************************************************
 *  \author
 *  	Rock Chen, BZ Robot INC
 *
 *  \version
 *	           BZROS_TASK V0.1
 *
 *	\par History
 *		- $log$
 *
 *	\par Release
 *		$Id: main.cpp$
 *		$Name:  $
 *  
 ****************************************************************************/

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#  include <unistd.h>
#  include <getopt.h>
#endif
#include <Bzr.hpp>
#include <BzrInterface.hpp>

void usage() {
    printf("RB[0..n]\n");
    printf("RB[0..n] jog_start <frame> <axis>\n");
    printf("RB[0..n] jog_end\n");
    printf("RB[0..n] speed [0..100]\n");
    printf("RB[0..n] PF pfType maxSpd maxAcc maxJerk Appr\n");  //PF is the command to set current profile, pfType=0 (trapzoidle type) =4 (s curve type)
                                                                //maxSpd,maxAcc,maxJerk are all in [0,1] representing the ratio of the ture maximal value of the robot (the true maximal value is configured through configuration utility software)
    printf("RB[0..n] FD baseno toolno ipo_mode\n"); //FD is the command to set current frame data, baseno in [0..7] is the current base no, toolno in [0..7] is the current tool no, 
                                                   // ipo_mode denotes the interpolation method used for orientation planning,
                                                   // ipo_mode is in [1..6], 1: using Euler angle w.r.t. base frame for IPO, 2: using Euler angle w.r.t. default tool frame for IPO, 3: using SingleAxis motion for IPO (any rotation is a rotation about a spatial axis based on Charles theorem)
                                                   // 5: usring Euler angle w.r.t. current tool frame for IPO
    /*
     * in the following three command lin, ptp,lins, there are needs to enter three flags about overhead, elbow (lefty or righty), and wrist (flip up or down)
     * bOH=0 means no overhead (normal), bOH=1 means overhead; bEB=0 means lefty (normal), bEB=1 means righty, bFL=0 means flip up, bFL=1 means flip down 
     */
    
    printf("RB[0..n] lin x y z A B C bOH bEB bFL\n"); // linear motion, target pos given as {x (mm),y (mm),z (mm), A (deg),B (deg),C (deg) bOH (int) bEB(int) bFL(int)}
    printf("RB[0..n] ptp x y z A B C bOH bEB bFL\n"); // PTP motion, target pos given as {x (mm),y (mm),z (mm), A (deg),B (deg),C (deg) bOH (int) bEB(int) bFL(int)}
    printf("RB[0..n] lins x y z A B C bOH bEB bFL spd \n"); //linear speed motino, target pos given as {x (mm),y (mm),z (mm), A (deg),B (deg),C (deg) bOH (int) bEB(int) bFL(int)}; target speed is given as spd
    printf("SO <chn> <val>\n");
    printf("IO 0..5\n");
    printf("GetDI 40..55");
    printf("SetDO 64..79");
    printf("read 0..7 -- print values of si/so/di/do/ai/ao/Axis/Pos. \n");
    printf("stop 0..1 -- stop remote services. \n");
    printf("config [1] -- print/save current config data. \n");
    printf("teach [1] -- print/save current teach data. \n");
    printf("prog [save] [xxx.lua] [xxxxxx] -- list/read/save user program. \n");
    printf("e -- switch to print errors or not. \n");
    printf("r -- switch to recovery mode or not. \n");
    printf("p -- print current position. \n");
    printf("q -- quit program. \n");
    printf("h -- for help. \n");
}

std::string str0, str1, str2, str3, str4;
int SystemStatusFun(void *userdata, int index, int value) {
    switch(index) {
    case 0:
        str0 = (value == 1) ? "" : "EMO";
        break;
    case 1:
        str1 = (value == 1) ? "" : "RESET";
        break;
    case 2:
        str2 = (value == 1) ? "AUTO" : (value == 2 ? "RECV" : (value == 3 ? "FREE" : "MAN"));
        break;
    case 3:
        str3 = (value == 1) ? "S_ON" : "S_OFF";
        break;
    case 4:
        switch(value) {
        case 0: str4 = "Stop"; break;
        case 1: str4 = "Run"; break;
        case 2: str4 = "Pause"; break;
        case 3: 
        default: str4 = "Cancel"; break;
        }
        break;
    }
    return 0;
}
int ProgressStatusFun(void *userdata, int value) {
    printf("Progress = %d\n", value);
    return 0;
}

int errPrint = 1;
int ErrorMessageFun(void *userdata, int code, const char*msg) {
    if (errPrint)
        printf("\nERROR (%d): %s\n", code, msg);
    return 0;
}

int ioPrint = -1;

EPos pos;
int UpdateCartFun(void *userdata, int index, const EPos *d) {
    pos = *d;
    if (ioPrint == 7)
        printf("x=%.2lf,y=%.2lf,z=%.2lf,A=%.2lf,B=%.2lf,C=%.2lf,S=%ld,T=%ld,E_A1=%.2lf,E_A2=%.2lf,E_A3=%.2lf\r",
            pos.x, pos.y, pos.z, pos.A, pos.B, pos.C, 
            pos.S, pos.T, pos.E_A1, pos.E_A2, pos.E_A3);
    return 0;
}
EAxis axis;
int UpdateJntFun(void *userdata, int index, const EAxis *d) {
    axis = *d;
    if (ioPrint == 6)
        printf("A1=%.2lf,A2=%.2lf,A3=%.2lf,A4=%.2lf,A5=%.2lf,A6=%.2lf,A7=%.2lf,A8=%.2lf,E_A1=%.2lf,E_A2=%.2lf,E_A3=%.2lf\r",
            axis.A1, axis.A2, axis.A3, axis.A4, axis.A5, axis.A6, 
            axis.A7, axis.A8, axis.E_A1, axis.E_A2, axis.E_A3);
    return 0;
}
void UpdateFun() {
    printf("x=%.2lf,y=%.2lf,z=%.2lf,A=%.2lf,B=%.2lf,C=%.2lf,S=%ld,T=%ld,E_A1=%.2lf,E_A2=%.2lf,E_A3=%.2lf\n",
        pos.x, pos.y, pos.z, pos.A, pos.B, pos.C, 
        pos.S, pos.T, pos.E_A1, pos.E_A2, pos.E_A3);
    printf("A1=%.2lf,A2=%.2lf,A3=%.2lf,A4=%.2lf,A5=%.2lf,A6=%.2lf,A7=%.2lf,A8=%.2lf,E_A1=%.2lf,E_A2=%.2lf,E_A3=%.2lf\n",
        axis.A1, axis.A2, axis.A3, axis.A4, axis.A5, axis.A6, 
        axis.A7, axis.A8, axis.E_A1, axis.E_A2, axis.E_A3);
}

int SafetyInFun(void *userdata, const char *name, int value) {
    if (ioPrint != 0) return 0;
    if (value) printf("%s = %d\n", name, value);
    return 0;
}
int SafetyOutFun(void *userdata, const char *name, int value) {
    if (ioPrint != 1) return 0;
    if (value) printf("%s = %d\n", name, value);
    return 0;
}
int DigitalInFun(void *userdata, const char *name, int value) {
    if (ioPrint != 2) return 0;
    if (value) printf("%s = %d\n", name, value);
    return 0;
}
int DigitalOutFun(void *userdata, const char *name, int value) {
    if (ioPrint != 3) return 0;
    if (value) printf("%s = %d\n", name, value);
    return 0;
}
int AnalogyInFun(void *userdata, const char *name, double value) {
    if (ioPrint != 4) return 0;
    printf("%s = %.2lf\n", name, value);
    return 0;
}
int AnalogyOutFun(void *userdata, const char *name, double value) {
    if (ioPrint != 5) return 0;
    printf("%s = %.2lf\n", name, value);
    return 0;
}

int PrintConfigData(int num_thread, const std::vector<std::string> &names) {
    printf("num_thread=%d\n", num_thread);
    for(unsigned i=0;i<names.size();i++)printf("%s\n",names[i].c_str());
    return 0;
}
int PrintTeachData(const ETeachData &d) {
//    for(int i=0;i<data.eaxis_var.size();i++)printf("%s\n",data.eaxis_var[i].c_str());
//    for(int i=0;i<data.epos_var.size();i++)printf("%s\n",data.epos_var[i].c_str());
//    for(int i=0;i<data.fdata_var.size();i++)printf("%s\n",data.fdata_var[i].c_str());
//    for(int i=0;i<data.profile_var.size();i++)printf("%s\n",data.profile_var[i].c_str());

    for(unsigned i=0;i<d.robotlist.size();i++) {
        printf("%s \n", d.robotlist[i].c_str());
        const EETeachData &data = d.rdata[i];
        for(unsigned i=0;i<data.epos.size();i++) {
            if (data.epos_var[i].length() == 0)continue; //ignore
            printf("%s= %f %f %f\n", data.epos_var[i].c_str(), data.epos[i].x, data.epos[i].y, data.epos[i].z);
        }
        for(unsigned i=0;i<data.eaxis.size();i++) {
            if (data.eaxis_var[i].length() == 0)continue; //ignore
            printf("%s= %f %f %f\n", data.eaxis_var[i].c_str(), data.eaxis[i].A1, data.eaxis[i].A2, data.eaxis[i].A3);
        }
        for(unsigned i=0;i<data.fdata.size();i++) {
            if (data.fdata_var[i].length() == 0)continue; //ignore
            printf("%s= %d %d %d %f\n", data.fdata_var[i].c_str(), 
                data.fdata[i].baseNO, data.fdata[i].toolNO, data.fdata[i].IPO_MODE, data.fdata[i].Circ_alpha);
        }
        for(unsigned i=0;i<data.profile.size();i++) {
            if (data.profile_var[i].length() == 0)continue; //ignore
            printf("%s= %d %f\n", data.profile_var[i].c_str(), data.profile[i].Type, data.profile[i].MaxJerk);
        }
        for(unsigned i=0;i<data.baseFrame.size();i++) {
            if (data.baseNames[i].length() == 0)continue; //ignore
            printf("%s= %f %f %f\n", data.baseNames[i].c_str(), data.baseFrame[i].x, data.baseFrame[i].y, data.baseFrame[i].z);
        }
        for(unsigned i=0;i<data.toolFrame.size();i++) {
            if (data.toolNames[i].length() == 0)continue; //ignore
            printf("%s= %f %f %f\n", data.toolNames[i].c_str(), data.toolFrame[i].x, data.toolFrame[i].y, data.toolFrame[i].z);
        }
    }
    
    return 0;
}
#ifndef WIN32
#  include <getopt.h> 
#else
#  include <getopt_win32.h>
#endif
int do_restart = 0;                 /* flag variables */
const char *ip = "localhost";       /* input IP address */
int iRecoveryMode = 0;

int argParse(int argc, char** argv) {
    struct option longopts[] = {
       { "nameserver",  optional_argument, NULL, 'n' },
       { "help",        no_argument,       NULL, 'h' },
       { 0, 0, 0, 0 }
    };

    int c;
    while ((c = getopt_long(argc, argv, ":hn:", longopts, NULL)) != -1) {
        switch (c) {
        case 'n':
            if (optarg != NULL)
               ip = optarg;
            break;
        case 'h':
            fprintf(stderr, "Usage: %s [<--nameserver | -n> [ip_address | localhost]] [--help | -h]\n",
                    argv[0]);
            return -1;
        case 0:     /* getopt_long() set a variable, just keep going */
            break;
#if 0
        case 1:
            /*
             * Use this case if getopt_long() should go through all
             * arguments. If so, add a leading '-' character to optstring.
             * Actual code, if any, goes here.
             */
            break;
#endif
        case ':':   /* missing option argument */
            fprintf(stderr, "%s: option '-%c' requires an argument\n",
                    argv[0], optopt);
            break;
        case '?':
        default:    /* invalid option */
            fprintf(stderr, "%s: option '-%c' is invalid: ignored\n",
                    argv[0], optopt);
            break;
        }
    }    
    return 0;
}

int main(int argc, char** argv) {
    int res = 0;
    std::string cmd;
    EventCallbacks cb = {0};
    HBZR h = NULL;
    EProfile cur_prof;
    //default current profile
    cur_prof.Type=4;
    cur_prof.MaxVel=0.5;
    cur_prof.MaxAcc=0.5;
    cur_prof.MaxJerk=0.5;
    cur_prof.AppDist=1.0;
    cur_prof.AccTime=0.1;
    cur_prof.DecTime=-1;
    EFrameData cur_fd;
    cur_fd.IPO_MODE=3; //rotation about a spatial axis for orientation ipo (interpolation)
    cur_fd.baseNO=0;
    cur_fd.toolNO=0;
    cur_fd.Circ_alpha=-1; //for circle motion with large angle of rotation
    cur_fd.Torque_state=false; //for future use
    
    
    res = argParse(argc, argv);
    if (res < 0) goto END;
    
    cb.status = SystemStatusFun;
    cb.prog = ProgressStatusFun;
    cb.err = ErrorMessageFun;
    cb.cart = UpdateCartFun;
    cb.jnt = UpdateJntFun;
    cb.sin = SafetyInFun;
    cb.sout = SafetyOutFun;
    cb.din = DigitalInFun;
    cb.dout = DigitalOutFun;
    cb.ain = AnalogyInFun;
    cb.aout = AnalogyOutFun;
    
    h = bzrCreate("./lib", cb);
    
    res = bzrConnect(h, ip, 5.0, 5.0);
    if (res < 0) { printf("Connect res = %d\n", res); goto END; }
    res = bzrStart(h, true);
    if (res < 0) { printf("Start callback res = %d\n", res); goto END; }
    
    while(1) {
        printf("%5s|%5s|%5s|%5s|%5s $ ", str0.c_str(), str1.c_str(), str2.c_str(), str3.c_str(), str4.c_str());
        std::getline(std::cin, cmd);
        if (cmd == "q" || cmd == "quit")
            break;
        else if (cmd == "e") {
            errPrint = errPrint == 1 ? 0 : 1;
            printf("Current errPrint = %d\n", errPrint);
        }else if (cmd == "h" || cmd == "help")
            usage();
        else if (cmd == "p" || cmd == "POS")
            UpdateFun();
        else if (cmd == "r") {
            iRecoveryMode = iRecoveryMode == 0 ? 1 : 0;
            res = bzrOperationModeSet(h, iRecoveryMode);
            printf("Current OperationMode = %d, res = %d\n", iRecoveryMode, res);
        } else if (cmd.substr(0,2) == "RB") {
            if (cmd.length() == 2) {
                int n = bzrRobotsNum(h);
                printf("Robots num = %d\n", n);
            } else {
                int i = strtol(cmd.substr(2).c_str(), NULL, 0);
                if (cmd.find("jog_start") != std::string::npos) {
                    int frame = 0, axis = 0;
                    std::size_t found = cmd.find("jog_start");
                    sscanf(cmd.substr(found).c_str(), "jog_start %d %d", &frame, &axis);
                    bzrRobotStartJog(h, i, frame, axis, -1);
                } else if (cmd.find("jog_end") != std::string::npos) {
                    bzrRobotEndJog(h, i);
                } else if (cmd.find("speed") != std::string::npos) {
                    std::size_t found = cmd.find("speed");
                    if (cmd.length() == found + 5) {
                        int speed = -1;
                        bzrRobotGetSpeed(h, i, speed);
                        printf("Robot %d speed = %d\n", i, speed);
                    } else {
                        int speed = strtol(cmd.substr(found+5).c_str(), NULL, 0);
                        bzrRobotSetSpeed(h, i, speed);
                    }
                }else if (cmd.find("lin") != std::string::npos){
                    std::size_t found = cmd.find("lin");
                    double x, y, z, A,B,C;
                    int bOH, bEB,bFL;
                    sscanf(cmd.substr(found).c_str(), "lin %lf %lf %lf %lf %lf %lf %d %d %d", &x, &y, &z, &A, &B, &C, &bOH,&bEB,&bFL);
                    
                    EPos pos;
                    pos.x=x;
                    pos.y=y;
                    pos.z=z;
                    pos.A=A;
                    pos.B=B;
                    pos.C=C;
                    pos.S=bOH+bEB*2+bFL*4;
                    pos.T=0; //default
                    pos.E_A1=0;
                    pos.E_A2=0;
                    pos.E_A3=0;
                   
                    
                    int res=bzrLinTo(h,i,pos, cur_prof, cur_fd); //again this just push the command to buffer, users still needs to servo on and then start the trajectory
                    if(res != 0){
                        printf("Failed to execute lin command (%d)\n", res);
                    }
                } 
                else if (cmd.find("ptp") != std::string::npos){
                    std::size_t found = cmd.find("ptp");
                    double x, y, z, A,B,C;
                    int bOH, bEB,bFL;
                    sscanf(cmd.substr(found).c_str(), "lin %lf %lf %lf %lf %lf %lf %d %d %d", &x, &y, &z, &A, &B, &C, &bOH,&bEB,&bFL);
                    
                    EPos pos;
                    pos.x=x;
                    pos.y=y;
                    pos.z=z;
                    pos.A=A;
                    pos.B=B;
                    pos.C=C;
                    pos.S=bOH+bEB*2+bFL*4;
                    pos.T=0; //default
                    pos.E_A1=0;
                    pos.E_A2=0;
                    pos.E_A3=0;
                   
                    
                    int res=bzrPtpTo(h,i,pos, cur_prof, cur_fd); //again this just push the command to buffer, users still needs to servo on and then start the trajectory
                    if(res != 0){
                        printf("Failed to execute Ptp command (%d)\n", res);
                    }
                } 
                else if (cmd.find("lins") != std::string::npos){
                    std::size_t found = cmd.find("lins");
                    double x, y, z, A,B,C, spd;
                    int bOH, bEB,bFL;
                    sscanf(cmd.substr(found).c_str(), "lins %lf %lf %lf %lf %lf %lf %d %d %d %lf", &x, &y, &z, &A, &B, &C, &bOH,&bEB,&bFL,&spd);
                    
                    EPos pos;
                    pos.x=x;
                    pos.y=y;
                    pos.z=z;
                    pos.A=A;
                    pos.B=B;
                    pos.C=C;
                    pos.S=bOH+bEB*2+bFL*4;
                    pos.T=0; //default
                    pos.E_A1=0;
                    pos.E_A2=0;
                    pos.E_A3=0;
                   
                    
                    int res=bzrLinSpdTo(h,i,pos, spd,cur_prof.AppDist, cur_fd); //again this just push the command to buffer, users still needs to servo on and then start the trajectory
                    if(res != 0){
                        printf("Failed to execute lins command (%d)\n", res);
                    }
                } 
                else {
                    std::string name;
                    RobotInfo rinfo;
                    bzrRobotInfo(h, i, rinfo);
                    name = rinfo.name;
                    printf("Robot %d name = %s\n", i, name.c_str());
                }
            }
        } else if (cmd.substr(0,2) == "SO") {
            int chn = 0, val = 0;
            sscanf(cmd.c_str(), "SO %d %d", &chn, &val);
            if (chn == -1)//Jump Safety In for simulation.
                bzrJumpSafeIn(h, val);
            else//Set Safety In/Out for simulation.
                bzrSetSafetyIO(h, chn, val);
        } else if (cmd.substr(0,2) == "IO") {
            IOChannels io = (IOChannels)strtol(cmd.substr(2).c_str(), NULL, 0);
            int n = bzrIOChannels(h, io);
            printf("IO %d num = %d\n", io, n);
        } else if (cmd.substr(0,4) == "read") {
            ioPrint = strtol(cmd.substr(4).c_str(), NULL, 0);
        } else if (cmd.substr(0,4) == "stop") {
            do_restart = strtol(cmd.substr(4).c_str(), NULL, 0);
            bzrStop(h, NULL, -1, do_restart);
        } else if (cmd.substr(0,5) == "teach") {
            ETeachData data;
            bzrTeachDataRead(h, data);
            PrintTeachData(data);
            if (cmd.length() > 6) {
                bzrTeachDataSave(h, data);
                printf("Teach Data saved.\n");
            }
        } else if (cmd.substr(0,6) == "config") {
            int num_thread = 0;
            std::vector<std::string> names;
            bzrConfigDataRead(h, num_thread, names);
            PrintConfigData(num_thread, names);
            if (cmd.length() > 7) {
                bzrConfigDataSave(h, num_thread, names);
                printf("Config Data saved.\n");
            }
        } else if (cmd.substr(0,4) == "prog") {
            if (cmd.length() < 6) {
                std::vector<std::string> names;
                bzrUserProgNamesRead(h, names);
                for (unsigned i=0; i < names.size(); i++)printf("%s\t", names[i].c_str());
                printf("\n");
            } else {
                if (cmd.substr(5,4) == "save") {  //prog save xxx.lua xxxxxx
                    if (cmd.find(" ", 10) != std::string::npos) {
                        std::size_t found = cmd.find(" ", 10);
                        std::string fname = cmd.substr(10, found-10);
                        std::string str = cmd.substr(found+1);
                        bzrUserProgSave(h, fname.c_str(), str);
                        printf("%s===\n%s\n", fname.c_str(), str.c_str());
                    }
                } else { //read
                    std::string str;
                    std::string fname = cmd.substr(5);
                    bzrUserProgRead(h, fname.c_str(), str);
                    printf("%s===\n%s\n", fname.c_str(), str.c_str());
                }
            }
        }
        else {
            ioPrint = -1; //disable print IO.
        }
    }
    
END:
    if(h)bzrDestroy(h);

    return res;
}
